/*
 * PHP Reformatter - A tool to make dirty PHP code readable.
 * Copyright (C) 2007 Joris van Rooij <jorrizza@jrrzz.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "preprocessor.h"

bool php_in_php;
bool php_preparse_got_error;
bool php_preparse_got_warning;
char global_buf[GLOBAL_BUFSIZE];
tag_t brackets[PREPROC_MAXDEPTH];
unsigned int depth;
unsigned int line;

size_t guess_strlen (char *code) {
  /* FIXME Fix uninitialized conditional jump in this function... Somewhere. */
  char prevbyte, startbyte;
  char stdoutbuf[512];
  unsigned short int stdoutbuflen;
  char *tmp;
  size_t stringlen = 1;

  /* Fix for ticket #1 - We handle stdout buffers as strings. */
  if (strncmp(code, "<<<", 3) == 0) {
    memset(stdoutbuf, 0, 512);
    stringlen = 3;
    code += 3;
    tmp = code;
    while (VALID_VARIABLE(*code)) {
      code++;
      stringlen++;
    }
    stdoutbuflen = code - tmp;
    strncpy(stdoutbuf, tmp, stdoutbuflen);
    while (strncmp(code, stdoutbuf, stdoutbuflen) != 0 && *code != 0x00) {
      if (*code == 0x0A)
        line++;
      code++;
      stringlen++;
    }
    return stringlen + stdoutbuflen;
  }

  if (*code != 0x22 && *code != 0x27 && *code != 0x60) /* We handle execution backticks as strings */
    return 0;

  startbyte = *code;

  code++;
  while (code != 0x00) {
    stringlen++;
    if (*code == 0x0A)
      line++;
    if ((*code == startbyte && prevbyte != 0x5C) || stringlen == GLOBAL_BUFSIZE)
      return stringlen;
    prevbyte = *code;
    code++;
  }

  return 0;
}

size_t guess_funclen (char *code) {
  size_t funclen = 0;

  while (VALID_FUNCTION(*code) && strlen(code) > 0 && funclen < GLOBAL_BUFSIZE) {
    funclen++;
    code++;
  }

  return funclen;
}

size_t guess_varlen (char *code) {
  size_t varlen = 1;

  if (*code != 0x24)
    return 0;

  code++;
  while (VALID_VARIABLE(*code) && strlen(code) > 0 && varlen < GLOBAL_BUFSIZE) {
    code++;
    varlen++;
  }

  return varlen;
}

size_t guess_deflen (char *code) {
  size_t deflen = 0;

  while (VALID_DEFINE(*code) && strlen(code) > 0 && deflen < GLOBAL_BUFSIZE) {
    code++;
    deflen++;
  }

  return deflen;
}

size_t guess_numlen (char *code) {
  size_t numlen = 1;
  bool had_special_char = FALSE;

  if (*code < 0x30 || *code > 0x39)
    return 0;

  code++;
  while ((VALID_NUMBER(*code) || *code == 0x78 || *code == 0x2E) && strlen(code) > 0 && numlen < GLOBAL_BUFSIZE) {
    if ((*code == 0x78 || *code == 0x2E) && !had_special_char)
      had_special_char = TRUE;
    else if ((*code == 0x78 || *code == 0x2E) && had_special_char)
      return numlen;
    code++;
    numlen++;
  }

  return numlen;
}

size_t guess_single_commentlen (char *code) {
  size_t commentlen = 0;

  while (*code != 0x0A && strlen(code) > 0 && commentlen < GLOBAL_BUFSIZE) {
    code++;
    commentlen++;
  }

  line++;
  return commentlen + 1; /* Also get newline */
}

size_t guess_multi_commentlen (char *code) {
  size_t commentlen = 0;

  if (strstr(code, "/*") != code)
    return 0;

  while (strstr(code, "*/") != code && strlen(code) > 0 && commentlen < GLOBAL_BUFSIZE) {
    if (*code == 0x0A)
      line++;

    code++;
    commentlen++;
  }

  return commentlen + 2;
}

size_t guess_undeflen (char *code) {
  size_t undeflen = 0;

  while (VALID_UNDEF(*code) && strlen(code) > 0 && undeflen < GLOBAL_BUFSIZE) {
    code++;
    undeflen++;
  }

  return undeflen;
}

tag_t guess_next_tag (char *code, tag_t prev, unsigned long int line, char *filename, bool loose) {
  /* This is kinda where the preprocessing magic happens */
  char buf[PREPROC_BUFSIZE], *t;
  size_t stringlen;
  php_preparse_got_warning = FALSE;
  php_preparse_got_error = FALSE;

  /* PHP_START_TAG */
  if ((strncmp(code, "<?php", strlen("<?php")) == 0 || strncmp(code, "<?PHP", strlen("<?PHP")) == 0) && !php_in_php) {
    php_in_php = TRUE;
    return PHP_START_TAG;
  }

  if (strncmp(code, "<?", strlen("<?")) == 0 && !php_in_php) {
    fprintf(stderr, "Preprocessor: %s: Warning: Incomplete PHP_START_TAG at line %lu.\n", filename, line);
    php_preparse_got_warning = TRUE;
    php_in_php = TRUE;
    return PHP_START_TAG;
  }

  /* PHP_END_TAG */
  if (strncmp(code, "?>", strlen("?>")) == 0 && php_in_php) {
    php_in_php = FALSE;
    return PHP_END_TAG;
  }

  /* PHP_COMMENT_SINGLE */
  if (strncmp(code, "//", strlen("//")) == 0 && php_in_php) {
    stringlen = guess_single_commentlen(code);
    memset(global_buf, 0, stringlen + 1);
    strncpy(global_buf, code, stringlen);
    return PHP_COMMENT_SINGLE;
  }
  if (strncmp(code, "#", strlen("#")) == 0 && php_in_php) {
    fprintf(stderr, "Preprocessor: %s: Warning: Use of Perl-like comment at line %lu.\n", filename, line);
    php_preparse_got_warning = TRUE;
    stringlen = guess_single_commentlen(code);
    memset(global_buf, 0, stringlen + 1);
    strncpy(global_buf, code, stringlen);
    return PHP_COMMENT_SINGLE;
  }

  /* PHP_COMMENT_MULTI */
  if (strncmp(code, "/*", strlen("/*")) == 0 && php_in_php) {
    stringlen = guess_multi_commentlen(code);
    memset(global_buf, 0, stringlen + 1);
    strncpy(global_buf, code, stringlen);
    if (strstr(global_buf + 2, "/*") != NULL) {
      fprintf(stderr, "Preprocessor: %s: Warning: Secondary comment tag in comment at line %lu.\n", filename, line);
      php_preparse_got_warning = TRUE;
    }
    return PHP_COMMENT_MULTI;
  }

  /* PHP_TYPECAST_INT */
  if (strncmp(code, "(int)", strlen("(int)")) == 0 && php_in_php)
    return PHP_TYPECAST_INT;

  /* PHP_TYPECAST_FLOAT */
  if (strncmp(code, "(float)", strlen("(float)")) == 0 && php_in_php)
    return PHP_TYPECAST_FLOAT;

  /* PHP_TYPECAST_STRING */
  if (strncmp(code, "(string)", strlen("(string)")) == 0 && php_in_php)
    return PHP_TYPECAST_STRING;

  /* PHP_TYPECAST_ARRAY */
  if (strncmp(code, "(array)", strlen("(array)")) == 0 && php_in_php)
    return PHP_TYPECAST_ARRAY;

  /* PHP_TYPECAST_OBJECT */
  if (strncmp(code, "(object)", strlen("(object)")) == 0 && php_in_php)
    return PHP_TYPECAST_OBJECT;

  /* PHP_FUNCTION_TYPE */
  if (strncmp(code, "function", strlen("function")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("function"))))
      return PHP_FUNCTION_TYPE;
  }

  /* PHP_FUNCTION_NAME */
  if (prev == PHP_FUNCTION_TYPE && php_in_php) {
    memset(global_buf, 0, guess_funclen(code) + 1);
    strncpy(global_buf, code, guess_funclen(code));
    if (strlen(global_buf) > 0)
      return PHP_FUNCTION_NAME;

    fprintf(stderr, "Preprocessor: %s: Error: Missing valid PHP_FUNCTION_NAME at line %lu.\n", filename, line);
    php_preparse_got_error = TRUE;
    return PHP_FUNCTION_NAME;
  }

  /* PHP_FUNCTION_BODY_START */
  if (prev == PHP_FUNCTION_NAME && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_FUNCTION_BODY_START;
    return PHP_FUNCTION_BODY_START;
  }

  /* PHP_FUNCTION_BODY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_FUNCTION_BODY_START) {
      depth--;
      return PHP_FUNCTION_BODY_END;
    }
  }

  /* PHP_FUNCTION_BODY_SEPERATOR */
  if (depth > 0 && strncmp(code, ",", strlen(",")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_FUNCTION_BODY_START)
      return PHP_FUNCTION_BODY_SEPERATOR;
  }

  /* PHP_FUNCTION_START */
  if (prev == PHP_FUNCTION_BODY_END && strncmp(code, "{", strlen("{")) == 0 && php_in_php) {
    brackets[depth++] = PHP_FUNCTION_START;
    return PHP_FUNCTION_START;
  }

  /* PHP_FUNCTION_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_FUNCTION_START) {
      depth--;
      return PHP_FUNCTION_END;
    }
  }

  /* PHP_LOOP_WHILE */
  if (strncmp(code, "while", strlen("while")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("while"))))
      return PHP_LOOP_WHILE;
  }

  /* PHP_LOOP_FOREACH */
  if (strncmp(code, "foreach", strlen("foreach")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("foreach"))))
      return PHP_LOOP_FOREACH;
  }

  /* PHP_LOOP_FOR */
  if (strncmp(code, "for", strlen("for")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("for"))))
      return PHP_LOOP_FOR;
  }

  /* PHP_LOOP_DO */
  if (strncmp(code, "do", strlen("do")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("do"))))
      return PHP_LOOP_DO;
  }

  /* PHP_LOOP_BODY_START */
  if ((prev == PHP_LOOP_FOR || prev == PHP_LOOP_FOREACH || prev == PHP_LOOP_WHILE) && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_LOOP_BODY_START;
    return PHP_LOOP_BODY_START;
  }

  /* PHP_LOOP_BODY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_LOOP_BODY_START) {
      depth--;
      return PHP_LOOP_BODY_END;
    }
  }

  /* PHP_LOOP_BODY_SEPERATOR */
  if (depth > 0 && strncmp(code, ";", strlen(";")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_LOOP_BODY_START)
      return PHP_LOOP_BODY_SEPERATOR;
  }

  /* PHP_LOOP_START */
  if ((prev == PHP_LOOP_BODY_END || prev == PHP_LOOP_DO) && strncmp(code, "{", strlen("{")) == 0 && php_in_php) {
    brackets[depth++] = PHP_LOOP_START;
    return PHP_LOOP_START;
  }

  /* PHP_LOOP_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_LOOP_START) {
      depth--;
      return PHP_LOOP_END;
    }
  }

  /* PHP_SWITCH_TYPE */
  if (strncmp(code, "switch", strlen("switch")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("switch"))))
      return PHP_SWITCH_TYPE;
  }

  /* PHP_SWITCH_BODY_START */
  if (prev == PHP_SWITCH_TYPE && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_SWITCH_BODY_START;
    return PHP_SWITCH_BODY_START;
  }

  /* PHP_SWITCH_BODY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_SWITCH_BODY_START) {
      depth--;
      return PHP_SWITCH_BODY_END;
    }
  }

  /* PHP_SWITCH_START */
  if (prev == PHP_SWITCH_BODY_END && strncmp(code, "{", strlen("{")) == 0 && php_in_php) {
    brackets[depth++] = PHP_SWITCH_START;
    return PHP_SWITCH_START;
  }

  /* PHP_SWITCH_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_SWITCH_START) {
      depth--;
      return PHP_SWITCH_END;
    }
  }

  /* PHP_SWITCH_CASE */
  if (depth > 0 && strncmp(code, ":", strlen(":")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_SWITCH_START)
      return PHP_SWITCH_CASE;
  }

  /* PHP_OOP_CLASS_TYPE */
  if (strncmp(code, "class", strlen("class")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("class")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_CLASS_TYPE) at line %lu.\n", filename, line);
      return PHP_OOP_CLASS_TYPE;
    }
  }

  /* PHP_OOP_CLASS_NAME */
  if ((prev == PHP_OOP_CLASS_TYPE || prev == PHP_OOP_KEYWORD_EXTENDS) && php_in_php && loose) {
    memset(global_buf, 0, guess_undeflen(code) + 1);
    strncpy(global_buf, code, guess_undeflen(code));
    if (strlen(global_buf) > 0) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_CLASS_NAME) at line %lu.\n", filename, line);
      return PHP_OOP_CLASS_NAME;
    }
  }

  /* PHP_OOP_CLASS_START */
  if (prev == PHP_OOP_CLASS_NAME && strncmp(code, "{", strlen("{")) == 0 && php_in_php && loose) {
    php_preparse_got_warning = TRUE;
    fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_CLASS_START) at line %lu.\n", filename, line);
    brackets[depth++] = PHP_OOP_CLASS_START;
    return PHP_OOP_CLASS_START;
  }

  /* PHP_OOP_CLASS_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php && loose) {
    if (brackets[depth - 1] == PHP_OOP_CLASS_START) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_CLASS_END) at line %lu.\n", filename, line);
      depth--;
      return PHP_OOP_CLASS_END;
    }
  }

  /* PHP_OOP_ARROW */
  if (strncmp(code, "->", strlen("->")) == 0 && php_in_php && loose) {
    php_preparse_got_warning = TRUE;
    fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_ARROW) at line %lu.\n", filename, line);
    return PHP_OOP_ARROW;
  }

  /* PHP_OOP_SCOPE_RESOLUTION */
  if (strncmp(code, "::", strlen("::")) == 0 && php_in_php && loose) {
    php_preparse_got_warning = TRUE;
    fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_SCOPE_RESOLUTION) at line %lu.\n", filename, line);
    return PHP_OOP_SCOPE_RESOLUTION;
  }

  /* PHP_ERROR_SILENCE */
  if (strncmp(code, "@", strlen("@")) == 0 && php_in_php)
    return PHP_ERROR_SILENCE;

  /* PHP_INCREMENTOR */
  if (strncmp(code, "++", strlen("++")) == 0 && php_in_php)
    return PHP_INCREMENTOR;

  /* PHP_DECREMENTOR */
  if (strncmp(code, "--", strlen("--")) == 0 && php_in_php)
    return PHP_DECREMENTOR;

  /* PHP_ARITHMETIC_NEGATION */
  if (prev == PHP_ASSIGNMENT && strncmp(code, "-", strlen("-")) == 0 && php_in_php)
    return PHP_ARITHMETIC_NEGATION;

  /* PHP_ARITHMETIC_NOT */
  if (prev == PHP_ASSIGNMENT && strncmp(code, "~", strlen("~")) == 0 && php_in_php)
    return PHP_ARITHMETIC_NOT;

  /* PHP_REFERENCE */
  if (strncmp(code, "&", strlen("&")) == 0 && php_in_php) {
    if (prev == PHP_ASSIGNMENT)
      return PHP_REFERENCE;
    else if (strncmp(code, "&$", strlen("&$")) == 0)
      return PHP_REFERENCE;
  }

  /* PHP_ARRAY_ARROW */
  if (strncmp(code, "=>", strlen("=>")) == 0 && php_in_php)
    return PHP_ARRAY_ARROW;

  /* PHP_STRING */
  if ((*code == 0x22 || *code == 0x27 || *code == 0x60 || strncmp(code, "<<<", strlen("<<<")) == 0) && php_in_php) {
    if ((stringlen = guess_strlen(code)) == 1) {
      fprintf(stderr, "Preprocessor: %s: Error: Invalid PHP_STRING at line %lu.\n", filename, line);
      php_preparse_got_error = TRUE;
      return PHP_STRING;
    }
    memset(global_buf, 0, stringlen + 1);
    strncpy(global_buf, code, stringlen);
    if (strstr(global_buf, "$") != NULL) {
      fprintf(stderr, "Preprocessor: %s: Warning: Possible unconcatenated variable(s) in string starting at line %lu.\n", filename, line);
      php_preparse_got_warning = TRUE;
    }
    return PHP_STRING;
  }

  /* PHP_COMPARISON, PHP_ARITHMETIC, PHP_ASSIGNMENT */
  if (strlen(code) >= strlen("===") && php_in_php) {
    memset(buf, 0, strlen("===") + 1);
    strncpy(buf, code, strlen("==="));
    if (strcmp(buf, "===") == 0 || strcmp(buf, "!==") == 0 || ((strcmp(buf, "and") == 0 || strcmp(buf, "xor") == 0) && !VALID_FUNCTION(*(code + strlen("and"))))) {
      memset(global_buf, 0, strlen("===") + 1);
      strncpy(global_buf, code, strlen("==="));
      return PHP_COMPARISON;
    }
    if (strcmp(buf, ">>=") == 0 || strcmp(buf, "<<=") == 0) {
      memset(global_buf, 0, strlen(">>=") + 1);
      strncpy(global_buf, code, strlen(">>="));
      return PHP_ASSIGNMENT;
    }
  }
  if (strlen(code) >= strlen("==") && php_in_php) {
    memset(buf, 0, strlen("==") + 1);
    strncpy(buf, code, strlen("=="));
    if (strcmp(buf, "==") == 0 || strcmp(buf, ">=") == 0 || strcmp(buf, "<=") == 0 || strcmp(buf, "<>") == 0 || strcmp(buf, "!=") == 0 || strcmp(buf, "||") == 0  || strcmp(buf, "&&") == 0 || (strcmp(buf, "or") == 0 && !VALID_FUNCTION(*(code + strlen("or"))))) {
      memset(global_buf, 0, strlen("==") + 1);
      strncpy(global_buf, code, strlen("=="));
      return PHP_COMPARISON;
    }
    if (strcmp(buf, ">>") == 0 || strcmp(buf, "<<") == 0) {
      memset(global_buf, 0, strlen(">>") + 1);
      strncpy(global_buf, code, strlen(">>"));
      return PHP_ARITHMETIC;
    }
    if (strcmp(buf, ".=") == 0 || strcmp(buf, "+=") == 0 || strcmp(buf, "-=") == 0 || strcmp(buf, "*=") == 0 || strcmp(buf, "/=") == 0 || strcmp(buf, "%=") == 0 || strcmp(buf, "&=") == 0 || strcmp(buf, "|=") == 0 || strcmp(buf, "^=") == 0) {
      memset(global_buf, 0, strlen(".=") + 1);
      strncpy(global_buf, code, strlen(".="));
      return PHP_ASSIGNMENT;
    }
  }
  if (strlen(code) >= strlen(">") && php_in_php) {
    memset(buf, 0, strlen(">") + 1);
    strncpy(buf, code, strlen(">"));
    if (strcmp(buf, ">") == 0 || strcmp(buf, "<") == 0) {
      memset(global_buf, 0, strlen(">") + 1);
      strncpy(global_buf, code, strlen(">"));
      return PHP_COMPARISON;
    }
    if (strcmp(buf, "+") == 0 || strcmp(buf, "-") == 0 || strcmp(buf, "*") == 0 || strcmp(buf, "/") == 0 || strcmp(buf, "%") == 0 || strcmp(buf, "^") == 0 || strcmp(buf, "|") == 0 || strcmp(buf, "&") == 0) {
      memset(global_buf, 0, strlen("+") + 1);
      strncpy(global_buf, code, strlen("+"));
      return PHP_ARITHMETIC;
    }
    if (strcmp(buf, "=") == 0) {
      memset(global_buf, 0, strlen("=") + 1);
      strncpy(global_buf, code, strlen("="));
      return PHP_ASSIGNMENT;
    }
  }

  /* PHP_ARITHMETIC_INV */
  if (strncmp(code, "!", strlen("!")) == 0 && php_in_php)
    return PHP_ARITHMETIC_INV;

  /* PHP_VARIABLE */
  if ((strncmp(code, "null", strlen("null")) == 0 || strncmp(code, "NULL", strlen("NULL")) == 0) && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("null")))) {
      memset(global_buf, 0, strlen("null") + 1);
      strncpy(global_buf, code, strlen("null"));
      return PHP_VARIABLE;
    }
  }
  if (*code == 0x24 && php_in_php) {
    if (!guess_varlen(code)) {
      fprintf(stderr, "Preprocessor: %s: Error: Invalid PHP_VARIABLE at line %lu.\n", filename, line);
      php_preparse_got_error = TRUE;
      return PHP_VARIABLE;
    }
    memset(global_buf, 0, guess_varlen(code) + 1);
    strncpy(global_buf, code, guess_varlen(code));
    return PHP_VARIABLE;
  }

  /* PHP_NUMBER */
  if (*code >= 0x30 && *code <= 0x39 && php_in_php) {
    if (!guess_numlen(code)) {
      fprintf(stderr, "Preprocessor: %s: Error: Invalid PHP_NUMBER at line %lu.\n", filename, line);
      php_preparse_got_error = TRUE;
      return PHP_NUMBER;
    }
    memset(global_buf, 0, guess_numlen(code) + 1);
    strncpy(global_buf, code, guess_numlen(code));
    return PHP_NUMBER;
  }

  /* PHP_ARRAY_TYPE */
  if (strncmp(code, "array", strlen("array")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("array"))))
      return PHP_ARRAY_TYPE;
  }

  /* PHP_ARRAY_START */
  if (prev == PHP_ARRAY_TYPE && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_ARRAY_START;
    return PHP_ARRAY_START;
  }

  /* PHP_ARRAY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_ARRAY_START) {
      depth--;
      return PHP_ARRAY_END;
    }
  }

  /* PHP_ARRAY_SEPERATOR */
  if (depth > 0 && strncmp(code, ",", strlen(",")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_ARRAY_START)
      return PHP_ARRAY_SEPERATOR;
  }

  /* PHP_IF_TYPE */
  if (strncmp(code, "if", strlen("if")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("if"))))
      return PHP_IF_TYPE;
  }

  /* PHP_IF_BODY_START */
  if ((prev == PHP_IF_TYPE || prev == PHP_ELSEIF_TYPE) && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_IF_BODY_START;
    return PHP_IF_BODY_START;
  }

  /* PHP_IF_BODY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_IF_BODY_START) {
      depth--;
      return PHP_IF_BODY_END;
    }
  }

  /* PHP_IF_START */
  if ((prev == PHP_IF_BODY_END || prev == PHP_ELSE_TYPE) && strncmp(code, "{", strlen("{")) == 0 && php_in_php) {
    brackets[depth++] = PHP_IF_START;
    return PHP_IF_START;
  }

  /* PHP_IF_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_IF_START) {
      depth--;
      return PHP_IF_END;
    }
  }

  /* PHP_ELSEIF_TYPE */
  if (strncmp(code, "elseif", strlen("elseif")) == 0 && php_in_php) {
    if(!VALID_FUNCTION(*(code + strlen("elseif"))))
      return PHP_ELSEIF_TYPE;
  }

  /* PHP_ELSE_TYPE */
  if (strncmp(code, "else", strlen("else")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("else"))))
      return PHP_ELSE_TYPE;
  }

  /* PHP_KEY_START */
  if ((prev == PHP_VARIABLE || prev == PHP_KEY_END || (prev == PHP_UNDEFINED && loose)) && strncmp(code, "[", strlen("[")) == 0 && php_in_php) {
    brackets[depth++] = PHP_KEY_START;
    return PHP_KEY_START;
  }

  /* PHP_KEY_END */
  if (depth > 0 && strncmp(code, "]", strlen("]")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_KEY_START) {
      depth--;
      return PHP_KEY_END;
    }
  }

  /* PHP_SUBSTRING_START */
  if ((prev == PHP_VARIABLE || (prev == PHP_UNDEFINED && loose)) && strncmp(code, "{", strlen("{")) == 0 && php_in_php) {
    brackets[depth++] = PHP_SUBSTRING_START;
    return PHP_SUBSTRING_START;
  }

  /* PHP_SUBSTRING_END */
  if (depth > 0 && strncmp(code, "}", strlen("}")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_SUBSTRING_START) {
      depth--;
      return PHP_SUBSTRING_END;
    }
  }

  /* PHP_CONCATENATOR */
  if (strncmp(code, ".", strlen(".")) == 0 && php_in_php)
    return PHP_CONCATENATOR;

  /* PHP_LINE_TERMINATOR */
  if (strncmp(code, ";", strlen(";")) == 0 && php_in_php)
    return PHP_LINE_TERMINATOR;

  /* PHP_FUNCTION_CALL */
  if ((stringlen = guess_funclen(code)) > 0 && php_in_php) {
    t = code + stringlen;
    while (strlen(t) > 0 && *t == 0x20)
      t++;
    if (*t == 0x28) {
      memset(global_buf, 0, stringlen + 1);
      strncpy(global_buf, code, stringlen);
      return PHP_FUNCTION_CALL;
    }
  }

  /* PHP_FUNCTION_CALL_BODY_START */
  if (prev == PHP_FUNCTION_CALL && strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_FUNCTION_CALL_BODY_START;
    return PHP_FUNCTION_CALL_BODY_START;
  }

  /* PHP_FUNCTION_CALL_BODY_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_FUNCTION_CALL_BODY_START) {
      depth--;
      return PHP_FUNCTION_CALL_BODY_END;
    }
  }

  /* PHP_FUNCTION_CALL_BODY_SEPERATOR */
  if (depth > 0 && strncmp(code, ",", strlen(",")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_FUNCTION_CALL_BODY_START)
      return PHP_FUNCTION_CALL_BODY_SEPERATOR;
  }

  /* PHP_BOOLGROUP_START */
  if (strncmp(code, "(", strlen("(")) == 0 && php_in_php) {
    brackets[depth++] = PHP_BOOLGROUP_START;
    return PHP_BOOLGROUP_START;
  }

  /* PHP_BOOLGROUP_END */
  if (depth > 0 && strncmp(code, ")", strlen(")")) == 0 && php_in_php) {
    if (brackets[depth - 1] == PHP_BOOLGROUP_START) {
      depth--;
      return PHP_BOOLGROUP_END;
    }
  }

  /* PHP_KEYWORD_BREAK */
  if (strncmp(code, "break", strlen("break")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("break"))))
      return PHP_KEYWORD_BREAK;
  }

  /* PHP_KEYWORD_EXIT */
  if (strncmp(code, "exit", strlen("exit")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("exit"))))
      return PHP_KEYWORD_EXIT;
  }

  /* PHP_KEYWORD_DIE */
  if (strncmp(code, "die", strlen("die")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("die"))))
      return PHP_KEYWORD_DIE;
  }

  /* PHP_KEYWORD_ECHO */
  if (strncmp(code, "echo", strlen("echo")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("echo"))))
      return PHP_KEYWORD_ECHO;
  }

  /* PHP_KEYWORD_PRINT */
  if (strncmp(code, "print", strlen("print")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("print"))))
      return PHP_KEYWORD_PRINT;
  }

  /* PHP_KEYWORD_REQUIREONCE */
  if (strncmp(code, "require_once", strlen("require_once")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("require_once"))))
      return PHP_KEYWORD_REQUIREONCE;
  }

  /* PHP_KEYWORD_REQUIRE */
  if (strncmp(code, "require", strlen("require")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("require"))))
      return PHP_KEYWORD_REQUIRE;
  }

  /* PHP_KEYWORD_INCLUDEONCE */
  if (strncmp(code, "include_once", strlen("include_once")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("include_once"))))
      return PHP_KEYWORD_INCLUDEONCE;
  }

  /* PHP_KEYWORD_INCLUDE */
  if (strncmp(code, "include", strlen("include")) == 0 && !VALID_FUNCTION(*(code + strlen("include"))) && php_in_php)
    return PHP_KEYWORD_INCLUDE;

  /* PHP_KEYWORD_RETURN */
  if (strncmp(code, "return", strlen("return")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("return"))))
      return PHP_KEYWORD_RETURN;
  }

  /* PHP_KEYWORD_CONTINUE */
  if (strncmp(code, "continue", strlen("continue")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("continue"))))
      return PHP_KEYWORD_CONTINUE;
  }

  /* PHP_KEYWORD_AS */
  if (strncmp(code, "as", strlen("as")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("as"))))
      return PHP_KEYWORD_AS;
  }

  /* PHP_KEYWORD_GLOBAL */
  if (strncmp(code, "global", strlen("global")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("global"))))
      return PHP_KEYWORD_GLOBAL;
  }

  /* PHP_KEYWORD_TRUE */
  if (strncmp(code, "true", strlen("true")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("true"))))
      return PHP_KEYWORD_TRUE;
  }

  /* PHP_KEYWORD_FALSE */
  if (strncmp(code, "false", strlen("false")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("false"))))
      return PHP_KEYWORD_FALSE;
  }

  /* PHP_KEYWORD_CASE */
  if (strncmp(code, "case", strlen("case")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("case"))))
      return PHP_KEYWORD_CASE;
  }

  /* PHP_KEYWORD_DEFAULT */
  if (strncmp(code, "default", strlen("default")) == 0 && php_in_php) {
    if (!VALID_FUNCTION(*(code + strlen("default"))))
      return PHP_KEYWORD_DEFAULT;
  }

  /* PHP_OOP_KEYWORD_NEW */
  if (strncmp(code, "new", strlen("new")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("new")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_NEW) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_NEW;
    }
  }

  /* PHP_OOP_KEYWORD_EXTENDS */
  if (strncmp(code, "extends", strlen("extends")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("extends")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_EXTENDS) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_EXTENDS;
    }
  }

  /* PHP_OOP_KEYWORD_PUBLIC */
  if (strncmp(code, "public", strlen("public")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("public")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_PUBLIC) at line %lu.\n", filename, line);
    }
    return PHP_OOP_KEYWORD_PUBLIC;
  }

  /* PHP_OOP_KEYWORD_PROTECTED */
  if (strncmp(code, "protected", strlen("protected")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("protected")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_PROTECTED) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_PROTECTED;
    }
  }

  /* PHP_OOP_KEYWORD_PRIVATE */
  if (strncmp(code, "private", strlen("private")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("private")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_PRIVATE) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_PRIVATE;
    }
  }

  /* PHP_OOP_KEYWORD_VAR */
  if (strncmp(code, "var", strlen("var")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("var")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_VAR) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_VAR;
    }
  }

  /* PHP_OOP_KEYWORD_CONST */
  if (strncmp(code, "const", strlen("const")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("const")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_CONST) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_CONST;
    }
  }

  /* PHP_OOP_KEYWORD_STATIC */
  if (strncmp(code, "static", strlen("static")) == 0 && php_in_php && loose) {
    if (!VALID_FUNCTION(*(code + strlen("static")))) {
      php_preparse_got_warning = TRUE;
      fprintf(stderr, "Preprocessor: %s: Warning: Found OOP code (PHP_OOP_KEYWORD_STATIC) at line %lu.\n", filename, line);
      return PHP_OOP_KEYWORD_STATIC;
    }
  }

  /* PHP_SHORTHAND_IF */
  if (strncmp(code, "?", strlen("?")) == 0 && php_in_php)
    return PHP_SHORTHAND_IF;

  /* PHP_SHORTHAND_ELSE */
  if (strncmp(code, ":", strlen(":")) == 0 && php_in_php)
    return PHP_SHORTHAND_ELSE;

  /* PHP_MISC_SEPERATOR */
  if (strncmp(code, ",", strlen(",")) == 0 && php_in_php)
    return PHP_MISC_SEPERATOR;

  /* PHP_STDOUT_DATA, PHP_DEFINED, PHP_UNDEFINED */
  if (php_in_php) {
    stringlen = guess_undeflen(code);
    if (loose && stringlen > 0) {
      memset(global_buf, 0, stringlen + 1);
      strncpy(global_buf, code, stringlen);
      if (guess_deflen(code) != stringlen) {
        php_preparse_got_warning = TRUE;
        fprintf(stderr, "Preprocessor: %s: Warning: Found PHP_UNDEFINED (%s) at line %lu.\n", filename, global_buf, line);
        fprintf(stderr, "Preprocessor: %s:          But allowing it in loose mode.\n", filename);
        return PHP_UNDEFINED;
      } else
        return PHP_DEFINED;
    }
    stringlen = guess_deflen(code);
    if (!loose && stringlen > 0) {
      memset(global_buf, 0, stringlen + 1);
      strncpy(global_buf, code, stringlen);
      return PHP_DEFINED;
    }
    fprintf(stderr, "Preprocessor: %s: Error: Parse error (unexpected '%c', 0x%02X) at line %lu.\n", filename, *code, *code, line);
    php_preparse_got_error = TRUE;
  }
  memset(global_buf, 0, 2);
  strncpy(global_buf, code, 1);
  return PHP_STDOUT_DATA;
}

int preprocess (char *code, size_t len, struct codeblock_s *codeblock, char *filename, bool loose) {
  size_t i = 0;
  tag_t prevtag = PHP_STDOUT_DATA;
  unsigned int warnings = 0;

  /* Reset all the globals in case we get called a second time */
  php_in_php = FALSE;
  php_preparse_got_error = FALSE;
  php_preparse_got_warning = FALSE;
  depth = 0;
  line = 1;

  while (i <= len) {
    while ((code[i] == 0x0A || code[i] == 0x0D || code[i] == 0x20 || code[i] == 0x09) && i < len - 1 && php_in_php) {
      if (code[i] == 0x0A)
        line++;
      i++;
    }
    if (code[i] == 0x0A && i < len - 1 && !php_in_php)
      line++;
    if (code[i] == 0x00)
      break;
    switch (guess_next_tag(&code[i], prevtag, line, filename, loose)) {
      case PHP_START_TAG:
        prevtag = PHP_START_TAG;
        codeblock_init(&codeblock->next, PHP_START_TAG, "<?php");
        codeblock = codeblock->next;
        if (php_preparse_got_warning)
          i += strlen("<?");
        else
          i += strlen("<?php");
        break;

      case PHP_END_TAG:
        prevtag = PHP_END_TAG;
        codeblock_init(&codeblock->next, PHP_END_TAG, "?>");
        codeblock = codeblock->next;
        i += strlen("?>");
        break;

      case PHP_COMMENT_SINGLE:
        /* prevtag = PHP_COMMENT_SINGLE; */
        codeblock_init(&codeblock->next, PHP_COMMENT_SINGLE, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_COMMENT_MULTI:
        /* prevtag = PHP_COMMENT_MULTI; */
        codeblock_init(&codeblock->next, PHP_COMMENT_MULTI, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_TYPECAST_INT:
        prevtag = PHP_TYPECAST_INT;
        codeblock_init(&codeblock->next, PHP_TYPECAST_INT, "(int)");
        codeblock = codeblock->next;
        i += strlen("(int)");
        break;

      case PHP_TYPECAST_FLOAT:
        prevtag = PHP_TYPECAST_FLOAT;
        codeblock_init(&codeblock->next, PHP_TYPECAST_FLOAT, "(float)");
        codeblock = codeblock->next;
        i += strlen("(float)");
        break;

      case PHP_TYPECAST_STRING:
        prevtag = PHP_TYPECAST_STRING;
        codeblock_init(&codeblock->next, PHP_TYPECAST_STRING, "(string)");
        codeblock = codeblock->next;
        i += strlen("(string)");
        break;

      case PHP_TYPECAST_ARRAY:
        prevtag = PHP_TYPECAST_ARRAY;
        codeblock_init(&codeblock->next, PHP_TYPECAST_ARRAY, "(array)");
        codeblock = codeblock->next;
        i += strlen("(array)");
        break;

      case PHP_TYPECAST_OBJECT:
        prevtag = PHP_TYPECAST_OBJECT;
        codeblock_init(&codeblock->next, PHP_TYPECAST_OBJECT, "(object)");
        codeblock = codeblock->next;
        i += strlen("(object)");
        break;

      case PHP_FUNCTION_TYPE:
        prevtag = PHP_FUNCTION_TYPE;
        codeblock_init(&codeblock->next, PHP_FUNCTION_TYPE, "function");
        codeblock = codeblock->next;
        i += strlen("function");
        break;

      case PHP_FUNCTION_NAME:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_FUNCTION_NAME;
        codeblock_init(&codeblock->next, PHP_FUNCTION_NAME, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_FUNCTION_BODY_START:
        prevtag = PHP_FUNCTION_BODY_START;
        codeblock_init(&codeblock->next, PHP_FUNCTION_BODY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_FUNCTION_BODY_END:
        prevtag = PHP_FUNCTION_BODY_END;
        codeblock_init(&codeblock->next, PHP_FUNCTION_BODY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_FUNCTION_BODY_SEPERATOR:
        prevtag = PHP_FUNCTION_BODY_SEPERATOR;
        codeblock_init(&codeblock->next, PHP_FUNCTION_BODY_SEPERATOR, ",");
        codeblock = codeblock->next;
        i += strlen(",");
        break;

      case PHP_FUNCTION_START:
        prevtag = PHP_FUNCTION_START;
        codeblock_init(&codeblock->next, PHP_FUNCTION_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_FUNCTION_END:
        prevtag = PHP_FUNCTION_END;
        codeblock_init(&codeblock->next, PHP_FUNCTION_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_LOOP_WHILE:
        prevtag = PHP_LOOP_WHILE;
        codeblock_init(&codeblock->next, PHP_LOOP_WHILE, "while");
        codeblock = codeblock->next;
        i += strlen("while");
        break;

      case PHP_LOOP_FOREACH:
        prevtag = PHP_LOOP_FOREACH;
        codeblock_init(&codeblock->next, PHP_LOOP_FOREACH, "foreach");
        codeblock = codeblock->next;
        i += strlen("foreach");
        break;

      case PHP_LOOP_FOR:
        prevtag = PHP_LOOP_FOR;
        codeblock_init(&codeblock->next, PHP_LOOP_FOR, "for");
        codeblock = codeblock->next;
        i += strlen("for");
        break;

      case PHP_LOOP_DO:
        prevtag = PHP_LOOP_DO;
        codeblock_init(&codeblock->next, PHP_LOOP_DO, "do");
        codeblock = codeblock->next;
        i += strlen("do");
        break;

      case PHP_LOOP_BODY_START:
        prevtag = PHP_LOOP_BODY_START;
        codeblock_init(&codeblock->next, PHP_LOOP_BODY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_LOOP_BODY_END:
        prevtag = PHP_LOOP_BODY_END;
        codeblock_init(&codeblock->next, PHP_LOOP_BODY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_LOOP_BODY_SEPERATOR:
        prevtag = PHP_LOOP_BODY_SEPERATOR;
        codeblock_init(&codeblock->next, PHP_LOOP_BODY_SEPERATOR, ";");
        codeblock = codeblock->next;
        i += strlen(";");
        break;

      case PHP_LOOP_START:
        prevtag = PHP_LOOP_START;
        codeblock_init(&codeblock->next, PHP_LOOP_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_LOOP_END:
        prevtag = PHP_LOOP_END;
        codeblock_init(&codeblock->next, PHP_LOOP_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_SWITCH_TYPE:
        prevtag = PHP_SWITCH_TYPE;
        codeblock_init(&codeblock->next, PHP_SWITCH_TYPE, "switch");
        codeblock = codeblock->next;
        i += strlen("switch");
        break;

      case PHP_SWITCH_BODY_START:
        prevtag = PHP_SWITCH_BODY_START;
        codeblock_init(&codeblock->next, PHP_SWITCH_BODY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_SWITCH_BODY_END:
        prevtag = PHP_SWITCH_BODY_END;
        codeblock_init(&codeblock->next, PHP_SWITCH_BODY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_SWITCH_START:
        prevtag = PHP_SWITCH_START;
        codeblock_init(&codeblock->next, PHP_SWITCH_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_SWITCH_END:
        prevtag = PHP_SWITCH_END;
        codeblock_init(&codeblock->next, PHP_SWITCH_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_SWITCH_CASE:
        prevtag = PHP_SWITCH_CASE;
        codeblock_init(&codeblock->next, PHP_SWITCH_CASE, ":");
        codeblock = codeblock->next;
        i += strlen(":");
        break;

      case PHP_OOP_CLASS_TYPE:
        prevtag = PHP_OOP_CLASS_TYPE;
        codeblock_init(&codeblock->next, PHP_OOP_CLASS_TYPE, "class");
        codeblock = codeblock->next;
        i += strlen("class");
        break;

      case PHP_OOP_CLASS_NAME:
        prevtag = PHP_OOP_CLASS_NAME;
        codeblock_init(&codeblock->next, PHP_OOP_CLASS_NAME, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_OOP_CLASS_START:
        prevtag = PHP_OOP_CLASS_START;
        codeblock_init(&codeblock->next, PHP_OOP_CLASS_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_OOP_CLASS_END:
        prevtag = PHP_OOP_CLASS_END;
        codeblock_init(&codeblock->next, PHP_OOP_CLASS_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_OOP_ARROW:
        prevtag = PHP_OOP_ARROW;
        codeblock_init(&codeblock->next, PHP_OOP_ARROW, "->");
        codeblock = codeblock->next;
        i += strlen("->");
        break;

      case PHP_OOP_SCOPE_RESOLUTION:
        prevtag = PHP_OOP_SCOPE_RESOLUTION;
        codeblock_init(&codeblock->next, PHP_OOP_SCOPE_RESOLUTION, "::");
        codeblock = codeblock->next;
        i += strlen("::");
        break;

      case PHP_ERROR_SILENCE:
        prevtag = PHP_ERROR_SILENCE;
        codeblock_init(&codeblock->next, PHP_ERROR_SILENCE, "@");
        codeblock = codeblock->next;
        i += strlen("@");
        break;

      case PHP_INCREMENTOR:
        prevtag = PHP_INCREMENTOR;
        codeblock_init(&codeblock->next, PHP_INCREMENTOR, "++");
        codeblock = codeblock->next;
        i += strlen("++");
        break;

      case PHP_DECREMENTOR:
        prevtag = PHP_DECREMENTOR;
        codeblock_init(&codeblock->next, PHP_DECREMENTOR, "--");
        codeblock = codeblock->next;
        i += strlen("--");
        break;

      case PHP_ARITHMETIC_NEGATION:
        prevtag = PHP_ARITHMETIC_NEGATION;
        codeblock_init(&codeblock->next, PHP_ARITHMETIC_NEGATION, "-");
        codeblock = codeblock->next;
        i += strlen("-");
        break;

      case PHP_ARITHMETIC_NOT:
        prevtag = PHP_ARITHMETIC_NOT;
        codeblock_init(&codeblock->next, PHP_ARITHMETIC_NOT, "~");
        codeblock = codeblock->next;
        i += strlen("~");
        break;

      case PHP_REFERENCE:
        prevtag = PHP_REFERENCE;
        codeblock_init(&codeblock->next, PHP_REFERENCE, "&");
        codeblock = codeblock->next;
        i += strlen("&");
        break;

      case PHP_ARRAY_ARROW:
        prevtag = PHP_ARRAY_ARROW;
        codeblock_init(&codeblock->next, PHP_ARRAY_ARROW, "=>");
        codeblock = codeblock->next;
        i += strlen("=>");
        break;

      case PHP_STRING:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_STRING;
        codeblock_init(&codeblock->next, PHP_STRING, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_ARITHMETIC:
        prevtag = PHP_ARITHMETIC;
        codeblock_init(&codeblock->next, PHP_ARITHMETIC, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_COMPARISON:
        prevtag = PHP_COMPARISON;
        codeblock_init(&codeblock->next, PHP_COMPARISON, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_ASSIGNMENT:
        prevtag = PHP_ASSIGNMENT;
        codeblock_init(&codeblock->next, PHP_ASSIGNMENT, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_ARITHMETIC_INV:
        prevtag = PHP_ARITHMETIC_INV;
        codeblock_init(&codeblock->next, PHP_ARITHMETIC_INV, "!");
        codeblock = codeblock->next;
        i += strlen("!");
        break;

      case PHP_VARIABLE:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_VARIABLE;
        codeblock_init(&codeblock->next, PHP_VARIABLE, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_NUMBER:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_NUMBER;
        codeblock_init(&codeblock->next, PHP_NUMBER, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_ARRAY_TYPE:
        prevtag = PHP_ARRAY_TYPE;
        codeblock_init(&codeblock->next, PHP_ARRAY_TYPE, "array");
        codeblock = codeblock->next;
        i += strlen("array");
        break;

      case PHP_ARRAY_START:
        prevtag = PHP_ARRAY_START;
        codeblock_init(&codeblock->next, PHP_ARRAY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_ARRAY_END:
        prevtag = PHP_ARRAY_END;
        codeblock_init(&codeblock->next, PHP_ARRAY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_ARRAY_SEPERATOR:
        prevtag = PHP_ARRAY_SEPERATOR;
        codeblock_init(&codeblock->next, PHP_ARRAY_SEPERATOR, ",");
        codeblock = codeblock->next;
        i += strlen(",");
        break;

      case PHP_IF_TYPE:
        prevtag = PHP_IF_TYPE;
        codeblock_init(&codeblock->next, PHP_IF_TYPE, "if");
        codeblock = codeblock->next;
        i += strlen("if");
        break;

      case PHP_IF_BODY_START:
        prevtag = PHP_IF_BODY_START;
        codeblock_init(&codeblock->next, PHP_IF_BODY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_IF_BODY_END:
        prevtag = PHP_IF_BODY_END;
        codeblock_init(&codeblock->next, PHP_IF_BODY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_IF_START:
        prevtag = PHP_IF_START;
        codeblock_init(&codeblock->next, PHP_IF_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_IF_END:
        prevtag = PHP_IF_END;
        codeblock_init(&codeblock->next, PHP_IF_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_ELSEIF_TYPE:
        prevtag = PHP_ELSEIF_TYPE;
        codeblock_init(&codeblock->next, PHP_ELSEIF_TYPE, "elseif");
        codeblock = codeblock->next;
        i += strlen("elseif");
        break;

      case PHP_ELSE_TYPE:
        prevtag = PHP_ELSE_TYPE;
        codeblock_init(&codeblock->next, PHP_ELSE_TYPE, "else");
        codeblock = codeblock->next;
        i += strlen("else");
        break;

      case PHP_KEY_START:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_KEY_START;
        codeblock_init(&codeblock->next, PHP_KEY_START, "[");
        codeblock = codeblock->next;
        i += strlen("[");
        break;

      case PHP_KEY_END:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_KEY_END;
        codeblock_init(&codeblock->next, PHP_KEY_END, "]");
        codeblock = codeblock->next;
        i += strlen("]");
        break;

      case PHP_SUBSTRING_START:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_SUBSTRING_START;
        codeblock_init(&codeblock->next, PHP_SUBSTRING_START, "{");
        codeblock = codeblock->next;
        i += strlen("{");
        break;

      case PHP_SUBSTRING_END:
        if (php_preparse_got_error)
          return 0;

        prevtag = PHP_SUBSTRING_END;
        codeblock_init(&codeblock->next, PHP_SUBSTRING_END, "}");
        codeblock = codeblock->next;
        i += strlen("}");
        break;

      case PHP_CONCATENATOR:
        prevtag = PHP_CONCATENATOR;
        codeblock_init(&codeblock->next, PHP_CONCATENATOR, ".");
        codeblock = codeblock->next;
        i += strlen(".");
        break;

      case PHP_LINE_TERMINATOR:
        prevtag = PHP_LINE_TERMINATOR;
        codeblock_init(&codeblock->next, PHP_LINE_TERMINATOR, ";");
        codeblock = codeblock->next;
        i += strlen(";");
        break;

      case PHP_FUNCTION_CALL:
        prevtag = PHP_FUNCTION_CALL;
        codeblock_init(&codeblock->next, PHP_FUNCTION_CALL, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_FUNCTION_CALL_BODY_START:
        prevtag = PHP_FUNCTION_CALL_BODY_START;
        codeblock_init(&codeblock->next, PHP_FUNCTION_CALL_BODY_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_FUNCTION_CALL_BODY_END:
        prevtag = PHP_FUNCTION_CALL_BODY_END;
        codeblock_init(&codeblock->next, PHP_FUNCTION_CALL_BODY_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_FUNCTION_CALL_BODY_SEPERATOR:
        prevtag = PHP_FUNCTION_CALL_BODY_SEPERATOR;
        codeblock_init(&codeblock->next, PHP_FUNCTION_CALL_BODY_SEPERATOR, ",");
        codeblock = codeblock->next;
        i += strlen(",");
        break;

      case PHP_BOOLGROUP_START:
        prevtag = PHP_BOOLGROUP_START;
        codeblock_init(&codeblock->next, PHP_BOOLGROUP_START, "(");
        codeblock = codeblock->next;
        i += strlen("(");
        break;

      case PHP_BOOLGROUP_END:
        prevtag = PHP_BOOLGROUP_END;
        codeblock_init(&codeblock->next, PHP_BOOLGROUP_END, ")");
        codeblock = codeblock->next;
        i += strlen(")");
        break;

      case PHP_KEYWORD_BREAK:
        prevtag = PHP_KEYWORD_BREAK;
        codeblock_init(&codeblock->next, PHP_KEYWORD_BREAK, "break");
        codeblock = codeblock->next;
        i += strlen("break");
        break;

      case PHP_KEYWORD_EXIT:
        prevtag = PHP_KEYWORD_EXIT;
        codeblock_init(&codeblock->next, PHP_KEYWORD_EXIT, "exit");
        codeblock = codeblock->next;
        i += strlen("exit");
        break;

      case PHP_KEYWORD_DIE:
        prevtag = PHP_KEYWORD_DIE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_DIE, "die");
        codeblock = codeblock->next;
        i += strlen("die");
        break;

      case PHP_KEYWORD_ECHO:
        prevtag = PHP_KEYWORD_ECHO;
        codeblock_init(&codeblock->next, PHP_KEYWORD_ECHO, "echo");
        codeblock = codeblock->next;
        i += strlen("echo");
        break;

      case PHP_KEYWORD_PRINT:
        prevtag = PHP_KEYWORD_PRINT;
        codeblock_init(&codeblock->next, PHP_KEYWORD_PRINT, "print");
        codeblock = codeblock->next;
        i += strlen("print");
        break;

      case PHP_KEYWORD_REQUIREONCE:
        prevtag = PHP_KEYWORD_REQUIREONCE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_REQUIREONCE, "require_once");
        codeblock = codeblock->next;
        i += strlen("require_once");
        break;

      case PHP_KEYWORD_REQUIRE:
        prevtag = PHP_KEYWORD_REQUIRE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_REQUIRE, "require");
        codeblock = codeblock->next;
        i += strlen("require");
        break;

      case PHP_KEYWORD_INCLUDEONCE:
        prevtag = PHP_KEYWORD_INCLUDEONCE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_INCLUDEONCE, "include_once");
        codeblock = codeblock->next;
        i += strlen("include_once");
        break;

      case PHP_KEYWORD_INCLUDE:
        prevtag = PHP_KEYWORD_INCLUDE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_INCLUDE, "include");
        codeblock = codeblock->next;
        i += strlen("include");
        break;

      case PHP_KEYWORD_RETURN:
        prevtag = PHP_KEYWORD_RETURN;
        codeblock_init(&codeblock->next, PHP_KEYWORD_RETURN, "return");
        codeblock = codeblock->next;
        i += strlen("return");
        break;

      case PHP_KEYWORD_CONTINUE:
        prevtag = PHP_KEYWORD_CONTINUE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_CONTINUE, "continue");
        codeblock = codeblock->next;
        i += strlen("continue");
        break;

      case PHP_KEYWORD_AS:
        prevtag = PHP_KEYWORD_AS;
        codeblock_init(&codeblock->next, PHP_KEYWORD_AS, "as");
        codeblock = codeblock->next;
        i += strlen("as");
        break;

      case PHP_KEYWORD_GLOBAL:
        prevtag = PHP_KEYWORD_GLOBAL;
        codeblock_init(&codeblock->next, PHP_KEYWORD_GLOBAL, "global");
        codeblock = codeblock->next;
        i += strlen("global");
        break;

      case PHP_KEYWORD_TRUE:
        prevtag = PHP_KEYWORD_TRUE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_TRUE, "true");
        codeblock = codeblock->next;
        i += strlen("true");
        break;

      case PHP_KEYWORD_FALSE:
        prevtag = PHP_KEYWORD_FALSE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_FALSE, "false");
        codeblock = codeblock->next;
        i += strlen("false");
        break;

      case PHP_KEYWORD_CASE:
        prevtag = PHP_KEYWORD_CASE;
        codeblock_init(&codeblock->next, PHP_KEYWORD_CASE, "case");
        codeblock = codeblock->next;
        i += strlen("case");
        break;

      case PHP_KEYWORD_DEFAULT:
        prevtag = PHP_KEYWORD_DEFAULT;
        codeblock_init(&codeblock->next, PHP_KEYWORD_DEFAULT, "default");
        codeblock = codeblock->next;
        i += strlen("default");
        break;

      case PHP_OOP_KEYWORD_NEW:
        prevtag = PHP_OOP_KEYWORD_NEW;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_NEW, "new");
        codeblock = codeblock->next;
        i += strlen("new");
        break;

      case PHP_OOP_KEYWORD_EXTENDS:
        prevtag = PHP_OOP_KEYWORD_EXTENDS;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_EXTENDS, "extends");
        codeblock = codeblock->next;
        i += strlen("extends");
        break;

      case PHP_OOP_KEYWORD_PUBLIC:
        prevtag = PHP_OOP_KEYWORD_PUBLIC;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_PUBLIC, "public");
        codeblock = codeblock->next;
        i += strlen("public");
        break;

      case PHP_OOP_KEYWORD_PROTECTED:
        prevtag = PHP_OOP_KEYWORD_PROTECTED;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_PROTECTED, "protected");
        codeblock = codeblock->next;
        i += strlen("protected");
        break;

      case PHP_OOP_KEYWORD_PRIVATE:
        prevtag = PHP_OOP_KEYWORD_PRIVATE;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_PRIVATE, "private");
        codeblock = codeblock->next;
        i += strlen("private");
        break;

      case PHP_OOP_KEYWORD_VAR:
        prevtag = PHP_OOP_KEYWORD_VAR;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_VAR, "var");
        codeblock = codeblock->next;
        i += strlen("var");
        break;

      case PHP_OOP_KEYWORD_CONST:
        prevtag = PHP_OOP_KEYWORD_CONST;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_CONST, "const");
        codeblock = codeblock->next;
        i += strlen("const");
        break;

      case PHP_OOP_KEYWORD_STATIC:
        prevtag = PHP_OOP_KEYWORD_STATIC;
        codeblock_init(&codeblock->next, PHP_OOP_KEYWORD_STATIC, "static");
        codeblock = codeblock->next;
        i += strlen("static");
        break;

      case PHP_SHORTHAND_IF:
        prevtag = PHP_SHORTHAND_IF;
        codeblock_init(&codeblock->next, PHP_SHORTHAND_IF, "?");
        codeblock = codeblock->next;
        i += strlen("?");
        break;

      case PHP_SHORTHAND_ELSE:
        prevtag = PHP_SHORTHAND_ELSE;
        codeblock_init(&codeblock->next, PHP_SHORTHAND_ELSE, ":");
        codeblock = codeblock->next;
        i += strlen(":");
        break;

      case PHP_MISC_SEPERATOR:
        prevtag = PHP_MISC_SEPERATOR;
        codeblock_init(&codeblock->next, PHP_MISC_SEPERATOR, ",");
        codeblock = codeblock->next;
        i += strlen(",");
        break;

      case PHP_DEFINED:
        prevtag = PHP_DEFINED;
        codeblock_init(&codeblock->next, PHP_DEFINED, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_UNDEFINED:
        prevtag = PHP_UNDEFINED;
        codeblock_init(&codeblock->next, PHP_UNDEFINED, global_buf);
        codeblock = codeblock->next;
        i += strlen(global_buf);
        break;

      case PHP_STDOUT_DATA:
        if (php_preparse_got_error)
          return 0;
        if (prevtag == PHP_STDOUT_DATA) {
          codeblock->content = realloc(codeblock->content, strlen(codeblock->content) + 2);
          strncat(codeblock->content, global_buf, 1);
        } else {
          prevtag = PHP_STDOUT_DATA;
          codeblock_init(&codeblock->next, PHP_STDOUT_DATA, global_buf);
          codeblock = codeblock->next;
        }
        i += 1;
        break;
    }
    if (php_preparse_got_warning) {
      warnings++;
      php_preparse_got_warning = FALSE;
    }
  }

  if (warnings > 0)
    fprintf(stderr, "Preprocessor: %s: Found %d warning(s).\n", filename, warnings);

  return 1;
}
